from hypothesis import given, strategies as st
import pydantic
import pytest
from typing_extensions import Annotated

from hrefs import Href, BaseReferrableModel, PrimaryKey
from util import hrefs


class Book(BaseReferrableModel):
    self: Annotated[Href["Book"], PrimaryKey(type_=int, name="id")]

    @classmethod
    def key_to_url(cls, key) -> str:
        path_params = cls.key_to_path_params(key)
        return f"/books/{path_params['id']}"

    @classmethod
    def url_to_key(cls, url: str) -> int:
        path_params = {"id": int(url.split("/")[-1])}
        key = cls.path_params_to_key(path_params)
        return key


class BookCover(BaseReferrableModel):
    book: Annotated[Href[Book], PrimaryKey]

    @classmethod
    def key_to_url(cls, key) -> str:
        path_params = cls.key_to_path_params(key)
        return f"/books/{path_params['book_id']}/cover"

    @classmethod
    def url_to_key(cls, url: str):
        parts = url.split("/")
        path_params = {"book_id": int(parts[-2])}
        return cls.path_params_to_key(path_params)


class Page(BaseReferrableModel):
    book: Annotated[Href[Book], PrimaryKey]
    page_number: Annotated[int, PrimaryKey]

    @classmethod
    def key_to_url(cls, key) -> str:
        path_params = cls.key_to_path_params(key)
        return f"/books/{path_params['book_id']}/pages/{path_params['page_number']}"

    @classmethod
    def url_to_key(cls, url: str):
        parts = url.split("/")
        path_params = {"book_id": int(parts[-3]), "page_number": int(parts[-1])}
        return cls.path_params_to_key(path_params)


class Bookmark(BaseReferrableModel):
    page: Annotated[Href[Page], PrimaryKey]

    @classmethod
    def key_to_url(cls, key) -> str:
        path_params = cls.key_to_path_params(key)
        return f"/books/{path_params['page_book_id']}/pages/{path_params['page_page_number']}/bookmark"

    @classmethod
    def url_to_key(cls, url: str):
        parts = url.split("/")
        path_params = {
            "page_book_id": int(parts[-4]),
            "page_page_number": int(parts[-2]),
        }
        return cls.path_params_to_key(path_params)


Book.update_forward_refs()

book_hrefs = hrefs(Book, st.integers())


@given(st.integers())
def test_self_href_from_key(key):
    book = Book(self=key)
    assert book.self == Href(key=key, url=Book.key_to_url(key))


@given(st.from_regex(r"\A/books/\d+\Z"))
def test_self_href_from_url(url):
    book = Book(self=url)
    assert book.self == Href(key=Book.url_to_key(url), url=url)


@given(book_hrefs)
def test_parse_href_key_from_referred_model(book_id):
    book = Book(self=book_id)
    href = pydantic.parse_obj_as(Href[BookCover], book)
    assert href.key == book.self
    assert href.url == BookCover.key_to_url(book.self)


@given(st.integers())
def test_parse_href_key_from_referred_key(book_id):
    href = pydantic.parse_obj_as(Href[BookCover], book_id)
    key = Href(key=book_id, url=Book.key_to_url(book_id))
    assert href.key == key
    assert href.url == BookCover.key_to_url(key)


@given(st.from_regex(r"\A/books/\d+/cover\Z"))
def test_parse_href_key_from_referred_url(url):
    href = pydantic.parse_obj_as(Href[BookCover], url)
    assert href.key == BookCover.url_to_key(url)
    assert href.url == url


@given(book_hrefs, st.integers())
def test_parse_composite_href_key_from_referred_model(book_id, page_number):
    book = Book(self=book_id)
    href = pydantic.parse_obj_as(Href[Page], (book, page_number))
    assert href.key == (book.self, page_number)
    assert href.url == Page.key_to_url((book.self, page_number))


@given(st.integers(), st.integers())
def test_parse_composite_href_key_from_referred_key(book_id, page_number):
    href = pydantic.parse_obj_as(Href[Page], (book_id, page_number))
    key = (Href(key=book_id, url=Book.key_to_url(book_id)), page_number)
    assert href.key == key
    assert href.url == Page.key_to_url(key)


@given(st.from_regex(r"\A/books/\d+/pages/\d+\Z"))
def test_parse_composite_href_key_from_referred_url(url):
    href = pydantic.parse_obj_as(Href[Page], url)
    assert href.key == Page.url_to_key(url)
    assert href.url == url


# Only real programmers use model keys that are hrefs to models that also have
# hrefs as model keys


@given(book_hrefs, st.integers())
def test_parse_indirect_href_key_from_referred_model(book_id, page_number):
    page = Page(book=book_id, page_number=page_number)
    href = pydantic.parse_obj_as(Href[Bookmark], page)
    assert href.key == pydantic.parse_obj_as(Href[Page], (page.book, page_number))
    assert href.url == Bookmark.key_to_url((page.book, page_number))


@given(st.integers(), st.integers())
def test_parse_indirect_href_key_from_referred_key(book_id, page_number):
    href = pydantic.parse_obj_as(Href[Bookmark], (book_id, page_number))
    page_key = (Href(key=book_id, url=Book.key_to_url(book_id)), page_number)
    key = Href(key=page_key, url=Page.key_to_url((book_id, page_number)))
    assert href.key == key
    assert href.url == Bookmark.key_to_url(key)


@given(st.from_regex(r"\A/books/\d+/pages/\d+/bookmark\Z"))
def test_parse_indirect_href_key_from_referred_url(url):
    href = pydantic.parse_obj_as(Href[Bookmark], url)
    assert href.key == Bookmark.url_to_key(url)
    assert href.url == url


def test_deep_indirection_is_not_supported():
    with pytest.raises(TypeError, match="not supported"):

        class SomeModelWithDeepIndirection(BaseReferrableModel):
            bookmark: Annotated[Href[Bookmark], PrimaryKey]



"""A test application

This is a self-contained demo of the `hrefs` library. You can run it with your
favorite ASGI server, and interact with it using your favorite HTTP client.

Its state is only contained in memory, so it is advisable also to manage your
libraries in a more persistent format!
"""

# pylint: disable=wrong-import-position,wrong-import-order

from typing import Dict, List
import uuid

from fastapi import FastAPI, HTTPException, Response
from fastapi.middleware import Middleware
from pydantic import BaseModel
from typing_extensions import Annotated

from hrefs import Href, PrimaryKey
from hrefs.starlette import ReferrableModel, HrefMiddleware


class _BookBase(BaseModel):
    title: str


class BookCreate(_BookBase):
    """Book creation payload"""


class Book(ReferrableModel, _BookBase):
    """Book in a library

    A book has a `title` and identity, represented by a hyperlink in the `self`
    field.
    """

    self: Annotated[Href["Book"], PrimaryKey(type_=uuid.UUID, name="id")]

    class Config:
        """Book config

        The `hrefs` library requires `details_view` option, which is the name of
        the Starlette route (in this case FastAPI route that is a Starlette
        route under the hood) serving the representation of a book.
        """

        details_view = "get_book"


Book.update_forward_refs()


class _LibraryBase(BaseModel):
    books: List[Href[Book]]


class LibraryCreate(_LibraryBase):
    """Library creation payload"""


class Library(ReferrableModel, _LibraryBase):
    """A library containing some books

    A library has a list of `books`, represented by hyperlinks. Its own identity
    in `self` is also a hyperlink.
    """

    self: Annotated[Href["Library"], PrimaryKey(type_=uuid.UUID, name="id")]

    class Config:
        """Book config

        The `hrefs` library requires `details_view` option, which is the name of
        the Starlette route (in this case FastAPI route that is a Starlette
        route under the hood) serving the representation of a library.
        """

        details_view = "get_library"


Library.update_forward_refs()

books: Dict[uuid.UUID, Book] = {}
libraries: Dict[uuid.UUID, Library] = {}

app = FastAPI(middleware=[Middleware(HrefMiddleware)])


@app.get("/libraries/{id}", response_model=Library)
def get_library(id: uuid.UUID):
    """Retrieve a library identified by `id`"""
    library = libraries.get(id)
    if not library:
        raise HTTPException(status_code=404, detail="Library not found")
    return library


@app.post("/libraries")
def post_library(library: LibraryCreate):
    """Create a new library

    A library can contain any number of books, represented by a list of
    hyperlinks. The books can either be referred by their `id` (an UUID encoded
    as a string), or a hyperlink (an URL previously returned by `POST /books`
    call).

    The identity will be automatically generated, and returned in the `Location`
    header.

    """
    if any(book.key not in books for book in library.books):
        raise HTTPException(
            status_code=400, detail="Trying to add nonexisting book to library"
        )
    new_library = Library(self=uuid.uuid4(), **library.dict())
    libraries[new_library.self.key] = new_library
    return Response(
        status_code=201,
        headers={"Location": new_library.self.url},
    )


@app.get("/books/{id}", response_model=Book)
def get_book(id: uuid.UUID):
    """Retrieve a book identified by `id`"""
    book = books.get(id)
    if not book:
        raise HTTPException(status_code=404, detail="Book not found")
    return book


@app.post("/books", status_code=204)
def post_book(book: BookCreate):
    """Create a new book

    The identity will be automatically generated, and returned in the `Location`
    header."""
    new_book = Book(self=uuid.uuid4(), **book.dict())
    books[new_book.self.key] = new_book
    return Response(status_code=201, headers={"Location": new_book.self.url})



"""Base models for referring and referrable types"""

import typing

import pydantic
import pydantic.typing
import typing_extensions

from .href import Href, Referrable

_DEFAULT_KEY = "id"


def _unwrap_key(obj: typing.Any):
    while isinstance(obj, Href):
        obj = obj.key
    return obj


def _getattr_and_maybe_unwrap_key(obj: typing.Any, name: str, unwrap_key: bool):
    ret = getattr(obj, name)
    if unwrap_key:
        ret = _unwrap_key(ret)
    return ret


class PrimaryKey:
    """Annotation declaring a field in :class:`BaseReferrableModel` as key

    ``PrimaryKey`` can be used the following way:

    .. code-block:: python

       from typing import Annotation

       class MyModel(BaseReferrableModel):
           my_id: Annotated[int, PrimaryKey]

           # ...the rest of the definitions...

    See :ref:`configure_key` for more details.

    Arguments:
        type_: The underlying key type if the annotated primary key is
               itself a hyperlink. See :ref:`href_as_key`.
        name: The name of the key. It may be distinct from the actual field
              name, and will appear in path parameters etc.
    """

    __slots__ = ["type_", "name"]

    def __init__(self, type_: typing.Type = None, name: str = None):
        self.type_ = type_
        self.name = name


# mypy doesn't like dynamic base classes:
# https://github.com/python/mypy/wiki/Unsupported-Python-Features
# ...so working around by some dirty casts
_base1: typing.Any = type(pydantic.BaseModel)
_base2: typing.Any = type(Referrable)


class _ReferrableModelKeyInfo(typing.NamedTuple):
    key_type: typing.Type
    should_unwrap_key: bool
    field_name: str


# pylint: disable=duplicate-bases,inconsistent-mro
class _ReferrableModelMeta(_base1, _base2):
    def __new__(cls, name, bases, namespace, **kwargs):
        annotations = pydantic.typing.resolve_annotations(
            namespace.get("__annotations__", {}), namespace.get("__module__", None)
        )
        key_names, key_infos = cls._create_key_names_and_types(name, annotations)
        assert len(key_names) == len(key_infos)

        if key_infos:
            if len(key_infos) > 1:
                (
                    key_model,
                    get_key,
                ) = cls._create_key_converters_multiple_types(key_names, key_infos)

            else:
                (
                    key_model,
                    get_key,
                ) = cls._create_key_converters_single_type(key_infos[0])

            namespace["_key_names"] = tuple(key_names)
            namespace["_key_model"] = key_model
            namespace["_get_key"] = get_key
            namespace["_key_map"] = {}

        return super().__new__(cls, name, bases, namespace, **kwargs)

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if hasattr(cls, "_key_model"):
            cls._calculate_key_map()

    @staticmethod
    def _create_key_names_and_types(
        name: str, all_annotations: typing.Mapping[str, typing.Any]
    ):
        key_names: typing.List[str] = []
        key_infos: typing.List[_ReferrableModelKeyInfo] = []
        for field_name, annotation in all_annotations.items():
            if typing_extensions.get_origin(annotation) is typing_extensions.Annotated:
                annotations = typing_extensions.get_args(annotation)
                key_annotations = [
                    key
                    for key in annotations
                    if key is PrimaryKey or isinstance(key, PrimaryKey)
                ]
                n_key_annotations = len(key_annotations)
                if n_key_annotations > 1:
                    raise TypeError(
                        f"{name}.{field_name}: Expected zero or one PrimaryKey annotations,"
                        f" got {n_key_annotations}"
                    )
                if n_key_annotations == 1:
                    origin_type = annotations[0]
                    primary_key_annotation: PrimaryKey = key_annotations[0]
                    # convert class into default instance
                    if primary_key_annotation is PrimaryKey:
                        primary_key_annotation = PrimaryKey()
                    type_from_annotation = primary_key_annotation.type_
                    key_name_from_annotation = primary_key_annotation.name
                    key_type = type_from_annotation or origin_type
                    should_unwrap_key = (
                        typing_extensions.get_origin(origin_type) is Href
                        and type_from_annotation is not None
                    )
                    key_names.append(key_name_from_annotation or field_name)
                    key_infos.append(
                        _ReferrableModelKeyInfo(key_type, should_unwrap_key, field_name)
                    )
        if not key_names and _DEFAULT_KEY in all_annotations:
            key_names.append(_DEFAULT_KEY)
            key_infos.append(
                _ReferrableModelKeyInfo(
                    all_annotations[_DEFAULT_KEY], False, _DEFAULT_KEY
                )
            )
        return key_names, key_infos

    @classmethod
    def _create_key_converters_multiple_types(
        cls,
        key_names: typing.Iterable[str],
        key_infos: typing.Iterable[_ReferrableModelKeyInfo],
    ):
        key_type = typing.NamedTuple(  # type: ignore
            "key",
            [
                (key_name, key_info.key_type)
                for (key_name, key_info) in zip(key_names, key_infos)
            ],
        )
        key_model = cls._create_key_model(key_type)

        def get_key(self):
            return key_model(
                __root__=[
                    _getattr_and_maybe_unwrap_key(
                        self, key_info.field_name, key_info.should_unwrap_key
                    )
                    for key_name, key_info in zip(key_names, key_infos)
                ]
            ).__root__

        return key_model, get_key

    @classmethod
    def _create_key_converters_single_type(cls, key_info: _ReferrableModelKeyInfo):
        key_model = cls._create_key_model(key_info.key_type)

        def get_key(self):
            return key_model(
                __root__=_getattr_and_maybe_unwrap_key(
                    self, key_info.field_name, key_info.should_unwrap_key
                )
            ).__root__

        return key_model, get_key

    @staticmethod
    def _create_key_model(key_type: typing.Type):
        return pydantic.create_model("key_model", __root__=(key_type, ...))


# pylint: disable=abstract-method
class BaseReferrableModel(
    pydantic.BaseModel, Referrable, metaclass=_ReferrableModelMeta
):
    """Referrable model with pydantic integration

    A subclass of both :class:`pydantic.BaseModel` and
    :class:`hrefs.Referrable`.  It should be used as the base class of any
    pydantic model that will be used as target of :class:`hrefs.Href`.

    ``BaseReferrableModel`` provides implementation on :func:`get_key()` and
    :func:`parse_as_key()` based on its field annotations. By default the model
    key is the ``id`` field (if it exists), but that can be changed by using
    :class:`PrimaryKey` to annotate other field(s).

    When using referrable models with FastAPI or Starlette in particular,
    :class:`hrefs.starlette.ReferrableModel` is more natural base.
    """

    _key_model: typing.ClassVar[typing.Type[pydantic.BaseModel]]
    _key_names: typing.ClassVar[typing.Tuple[str, ...]]
    _get_key: typing.ClassVar[typing.Callable[["BaseReferrableModel"], typing.Any]]
    _key_map: typing.ClassVar[
        typing.Dict[str, typing.Union[str, typing.Tuple[str, ...]]]
    ]

    def get_key(self) -> typing.Any:
        """Return the model key

        Returns:
            The model key based on the field annotations. If the key is
            composite, return a tuple containing the parts.
        """
        return self._get_key()

    @classmethod
    def has_simple_key(cls) -> bool:
        """Query if the model has a simple key

        Returns:
            ``True`` if the model has simple (single part) key, ``False``
            otherwise
        """
        return len(cls._key_names) == 1

    @classmethod
    def key_to_path_params(cls, key: typing.Any) -> typing.Dict[str, typing.Any]:
        """Convert model key to path parameters

        This is a helper that can be used to convert a model key into a
        dictionary containing the key parts. Hyperlinks are unwrapped (see
        :ref:`href_as_key`).  It can be used to generate URLs in several HTTP
        frameworks.

        Arguments:
            key: model key

        Returns:
            A dictionary mapping key names to key parts

        """
        if cls.has_simple_key():
            key = (key,)
        path_params = {}
        for subkeys, subkey_names in zip(key, cls._key_map.values()):
            subkeys = _unwrap_key(subkeys)
            if isinstance(subkey_names, str):
                path_params[subkey_names] = subkeys
            else:
                for subkey_name, subkey in zip(subkey_names, subkeys):
                    path_params[subkey_name] = subkey
        return path_params

    @classmethod
    def path_params_to_key(
        cls, path_params: typing.Mapping[str, typing.Any]
    ) -> typing.Any:
        """Convert path parameters to model key

        This helper can be used to convert path parameter mapping to model
        key. It is the inverse of :meth:`key_to_path_params()`.

        Arguments:
            path_params: A mapping from key names to key parts

        Returns:
            Model key parsed from ``path_params``
        """
        subkeys = []
        for subkey_names in cls._key_map.values():
            if isinstance(subkey_names, str):
                subkey = path_params[subkey_names]
            else:
                subkey = [path_params[subkey_name] for subkey_name in subkey_names]
            subkeys.append(subkey)
        if cls.has_simple_key():
            subkeys = subkeys[0]
        return cls.parse_as_key(subkeys)

    @staticmethod
    def try_parse_as(
        model: typing.Type[pydantic.BaseModel], value: typing.Any
    ) -> typing.Optional[typing.Any]:
        """Parse ``value`` as ``model`` but swallow validation error

        Arguments:
            model: the model parsed to
            value: the value to parse

        Returns:
            ``value`` parsed as ``model``, or ``None`` on validation error
        """
        try:
            return model.parse_obj(value).__root__
        except pydantic.ValidationError:
            return None

    @classmethod
    def parse_as_key(cls, value: typing.Any) -> typing.Optional[typing.Any]:
        """Parse ``value`` as the key type

        The type of the model key based on the field annotations. Either a
        single type, or (in case of composite key), a tuple of the parts.
        """
        return cls.try_parse_as(cls._key_model, value)

    @classmethod
    def update_forward_refs(cls, **localns: typing.Any) -> None:
        super().update_forward_refs(**localns)
        cls._key_model.update_forward_refs(**localns)
        cls._calculate_key_map()

    @classmethod
    def _calculate_key_map(cls):
        key_type = cls._key_model.__fields__["__root__"].outer_type_
        key_types: typing.Dict[str, typing.Type]
        if cls.has_simple_key():
            key_name = cls._key_names[0]
            key_types = {key_name: key_type}
        else:
            key_types = key_type.__annotations__

        cls._key_map.clear()
        for key_name, key_type in key_types.items():
            target_key_name = key_name
            # If key part is `Href`, we examine the target and unwrap it
            if typing_extensions.get_origin(key_type) is Href:
                (target_type,) = typing_extensions.get_args(key_type)
                target_type_key_map = getattr(target_type, "_key_map", None)
                if target_type_key_map:
                    target_type_key_names = target_type_key_map.values()
                    # This would get complicated: if the target of `Href` also
                    # has complex key that needs unwrapping (indirection two
                    # levels deep), we don't do that. It would be possible if
                    # calculating key map was properly recursive, though.
                    if not all(isinstance(name, str) for name in target_type_key_names):
                        raise TypeError(
                            "Href to models with complex key are not supported as model key. "
                            f"{target_type!r} has key map {target_type_key_map!r}"
                        )
                    target_key_name = [
                        f"{key_name}_{target_key_name}"
                        for target_key_name in target_type_key_map.values()
                    ]
                    target_key_name = (
                        tuple(target_key_name)
                        if len(target_key_name) > 1
                        else target_key_name[0]
                    )
            cls._key_map[key_name] = target_key_name



from typing import List, Tuple

from hypothesis import given, strategies as st
import pydantic
from typing_extensions import Annotated

from hrefs import Href, BaseReferrableModel, PrimaryKey


class Book(BaseReferrableModel):
    id: int

    @staticmethod
    def key_to_url(key: int) -> str:
        return f"/books/{key}"

    @staticmethod
    def url_to_key(url: str):
        return int(url.split("/")[-1])


class Page(BaseReferrableModel):
    book_id: Annotated[int, PrimaryKey]
    page_number: Annotated[int, PrimaryKey]

    @staticmethod
    def key_to_url(key: Tuple[int, int]) -> str:
        return f"/books/{key[0]}/pages/{key[1]}"

    @staticmethod
    def url_to_key(url: str):
        parts = url.split("/")
        return int(parts[-3]), int(parts[-1])


@given(st.builds(Page))
def test_parse_model_with_composite_key_to_href(page):
    href = pydantic.parse_obj_as(Href[Page], page)
    key = (page.book_id, page.page_number)
    assert href.key == key
    assert href.url == Page.key_to_url(key)


@given(st.tuples(st.integers(), st.integers()))
def test_parse_composite_key_to_href(key):
    href = pydantic.parse_obj_as(Href[Page], key)
    assert href.key == key
    assert href.url == Page.key_to_url(key)


@given(st.from_regex(r"\A/books/\d+/pages/\d+\Z"))
def test_parse_composite_url_to_href(url):
    href = pydantic.parse_obj_as(Href[Page], url)
    assert href.key == Page.url_to_key(url)
    assert href.url == url



import json
import typing

from hypothesis import given, strategies as st
import pydantic
import pytest
from typing_extensions import Annotated

from hrefs import Href, BaseReferrableModel, PrimaryKey
from util import hrefs


class Pet(BaseReferrableModel):
    id: int

    @staticmethod
    def key_to_url(key: int) -> str:
        return f"/pets/{key}"

    @staticmethod
    def url_to_key(url: str):
        return int(url.split("/")[-1])


class Owner(pydantic.BaseModel):
    pets: typing.List[Href[Pet]]


pet_hrefs = hrefs(Pet, st.integers())


@given(pet_hrefs)
def test_parse_href(href):
    assert pydantic.parse_obj_as(Href[Pet], href) is href


@given(st.builds(Pet))
def test_parse_referrable_model(pet):
    href = pydantic.parse_obj_as(Href[Pet], pet)
    assert href.key == pet.id
    assert href.url == Pet.key_to_url(pet.id)


@given(st.integers())
def test_parse_key_to_href(key):
    href = pydantic.parse_obj_as(Href[Pet], key)
    assert href.key == key
    assert href.url == Pet.key_to_url(key)


@given(st.from_regex(r"\A/pets/\d+\Z"))
def test_parse_url_to_key(url):
    href = pydantic.parse_obj_as(Href[Pet], url)
    assert href.key == Pet.url_to_key(url)
    assert href.url == url


def test_parse_href_with_unparseable_key_fails():
    with pytest.raises(pydantic.ValidationError):
        pydantic.parse_obj_as(Href, object())


def test_parse_href_without_parameter_fails():
    with pytest.raises(pydantic.ValidationError):
        pydantic.parse_obj_as(Href, 123)


@given(st.builds(Owner, pets=st.lists(pet_hrefs)))
def test_json_encode(owner):
    owner_json = json.loads(owner.json())
    assert owner_json["pets"] == [pet.url for pet in owner.pets]


@given(st.integers())
def test_primary_key_annotation(my_id):
    class MyModel(BaseReferrableModel):
        my_id: Annotated[int, PrimaryKey]

        @staticmethod
        def key_to_url(key: int) -> None:
            ...

        @staticmethod
        def url_to_key(url: str):
            ...

    assert MyModel(my_id=my_id).get_key() == my_id


def test_multiple_primary_key_annotations_fails():
    with pytest.raises(TypeError):

        class MyModel(BaseReferrableModel):
            my_id: Annotated[int, PrimaryKey, PrimaryKey]

            @staticmethod
            def key_to_url(key: int) -> None:
                ...

            @staticmethod
            def url_to_key(url: str):
                ...


def test_href_forward_reference():
    class MyModel(BaseReferrableModel):
        id: int
        self: Href["MyModel"]

        @pydantic.root_validator(pre=True)
        def populate_self(cls, values):
            values["self"] = values["id"]
            return values

        @staticmethod
        def key_to_url(key: int) -> None:
            return f"/{key}"

        @staticmethod
        def url_to_key(url: str):
            ...

    MyModel.update_forward_refs()

    assert MyModel(id=1).self == Href(key=1, url="/1")



import contextvars
import json
import random
import typing
import uuid

import fastapi
import fastapi.middleware
import fastapi.testclient
from hypothesis import given, strategies as st
import pydantic
from typing_extensions import Annotated

from hrefs import Href, PrimaryKey
from hrefs.starlette import ReferrableModel, HrefMiddleware


class Comment(ReferrableModel):
    id: uuid.UUID

    class Config:
        details_view = "get_comment"


class Article(ReferrableModel):
    self: Annotated[Href["Article"], PrimaryKey(type_=uuid.UUID, name="id")]
    comments: typing.List[Href[Comment]]
    current_revision: Href["ArticleRevision"]

    class Config:
        details_view = "get_article"


class ArticleRevision(ReferrableModel):
    article: Annotated[Href[Article], PrimaryKey]
    revision: Annotated[int, PrimaryKey]

    class Config:
        details_view = "get_revision"


Article.update_forward_refs()


article_var: contextvars.ContextVar[uuid.UUID] = contextvars.ContextVar("article_var")
revision_var: contextvars.ContextVar[int] = contextvars.ContextVar("revision_var")
comments_var: contextvars.ContextVar[typing.List[uuid.UUID]] = contextvars.ContextVar(
    "comments_var"
)
save_article_var: contextvars.ContextVar[
    typing.Callable[[Article], None]
] = contextvars.ContextVar("save_article_var")


app = fastapi.FastAPI(middleware=[fastapi.middleware.Middleware(HrefMiddleware)])


@app.get("/articles/{id}", response_model=Article)
async def get_article(id: uuid.UUID):
    assert id == article_var.get()
    return Article(
        self=id,
        comments=comments_var.get(),
        current_revision=(id, revision_var.get()),
    )


@app.post("/articles")
async def post_article(article: Article):
    save_article = save_article_var.get()
    save_article(article)


@app.get("/comments/{id}", response_model=Comment)
async def get_comment(id: uuid.UUID):
    assert id in comments_var.get()
    return Comment(id=id)


@app.get("/articles/{article_id}/revisions/{revision}")
async def get_revision(article_id: uuid.UUID, revision: int):
    assert article_id == article_var.get()
    return ArticleRevision(article=article_id, revision=revision)


client = fastapi.testclient.TestClient(app)


@given(st.uuids(), st.integers(), st.lists(st.uuids(), min_size=1))
def test_parse_key_to_href(article_id, revision, comment_ids):
    article_var.set(article_id)
    revision_var.set(revision)
    comments_var.set(comment_ids)
    response = client.get(f"/articles/{article_id}")
    assert response.status_code == fastapi.status.HTTP_200_OK, response.text
    article = response.json()
    comment_href, comment_id = random.choice(
        list(zip(article["comments"], comment_ids))
    )
    comment = client.get(comment_href).json()
    assert uuid.UUID(comment["id"]) == comment_id
    current_revision_href = article["current_revision"]
    current_revision = client.get(current_revision_href).json()
    assert current_revision == {
        "article": f"http://testserver/articles/{article_id}",
        "revision": revision,
    }


@given(st.uuids(), st.integers(), st.lists(st.uuids(), min_size=1))
def test_parse_url_to_href(article_id, revision, comment_ids):
    def assert_article(article: Article):
        assert article.self.key == article_id
        assert [comment_href.key for comment_href in article.comments] == comment_ids
        _article, _revision = article.current_revision.key
        assert _article.key == article_id
        assert _revision == revision

    save_article_var.set(assert_article)
    response = client.post(
        "/articles",
        data=json.dumps(
            dict(
                self=str(article_id),
                comments=[f"http://testclient/comments/{id}" for id in comment_ids],
                current_revision=f"http://testclient/articles/{article_id}/revisions/{revision}",
            )
        ),
    )
    assert response.status_code == fastapi.status.HTTP_200_OK, response.text


@given(st.uuids(), st.lists(st.uuids(), min_size=1))
def test_parse_invalid_url_fails(article_id, comment_ids):
    response = client.post(
        "/articles",
        data=json.dumps(
            dict(
                self=str(article_id),
                comments=[
                    f"http://testclient/not/a/real/route/{id}" for id in comment_ids
                ],
            )
        ),
    )
    assert response.status_code == fastapi.status.HTTP_422_UNPROCESSABLE_ENTITY



"""Starlette integration"""

import contextvars
import typing

import pydantic
import starlette.requests
import starlette.middleware.base

from .model import BaseReferrableModel


_URL_MODEL: typing.Type[pydantic.BaseModel] = pydantic.create_model(
    "_URL_MODEL", __root__=(pydantic.AnyHttpUrl, ...)
)

_request_var: contextvars.ContextVar[
    starlette.requests.Request
] = contextvars.ContextVar("_request_var")


class HrefMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """Middleware for resolving hrefs

    This middleware needs to be added to the middleware stack of the Starlette
    app using the :class:`ReferrableModel`.
    """

    @staticmethod
    async def dispatch(request: starlette.requests.Request, call_next):
        _request_var.set(request)
        return await call_next(request)


# pylint: disable=abstract-method
class ReferrableModel(BaseReferrableModel):
    """Referrable model with Starlette integration

    This class implements the :class:`hrefs.BaseReferrableModel` with the
    following features:

    * Its key type is inferred from type annotations as in the base class

    * Its URL type is always :class:`pydantic.AnyHttpUrl`

    * When converting between key and URL, it relies on the Starlette request
      objects. Thus a :class:`ReferrableModel` is only usable within request
      handlers or middleware above the :class:`HrefMiddleware` in the middleware
      stack of the application.

    Here is a minimal example using route called ``"my_view"`` to convert
    to/from URLs:

    .. code-block:: python

        class MyModel(ReferrableModel):
            id: int

            class Config:
                details_view = "my_view"

    For a more complete example of using :mod:`hrefs` library with Starlette,
    please refer to :ref:`quickstart`.
    """

    @classmethod
    def parse_as_url(cls, value: typing.Any) -> typing.Optional[pydantic.AnyHttpUrl]:
        """Parse ``value`` as ``pydantic.AnyHttpUrl``"""
        return cls.try_parse_as(_URL_MODEL, value)

    @classmethod
    def key_to_url(cls, key) -> pydantic.AnyHttpUrl:
        request = _request_var.get()
        details_view = cls._get_details_view()
        kwargs = cls.key_to_path_params(key)
        return pydantic.parse_obj_as(
            pydantic.AnyHttpUrl, request.url_for(details_view, **kwargs)
        )

    @classmethod
    def url_to_key(cls, url: pydantic.AnyHttpUrl) -> typing.Any:
        request = _request_var.get()
        details_view = cls._get_details_view()
        for route in request.app.routes:
            if route.name == details_view:
                _, scope = route.matches(
                    {"type": "http", "method": "GET", "path": url.path}
                )
                if scope:
                    return cls.path_params_to_key(scope["path_params"])
        raise ValueError(f"Could not resolve {url} into key")

    @classmethod
    def _get_details_view(cls):
        return getattr(cls.__config__, "details_view")
